@@ -1,11 +1,11 @@
 /* System dependent definitions for GNU tar.
 
    Copyright (C) 1994, 1995, 1996, 1997, 1998, 1999, 2000, 2001, 2003,
-   2004, 2005, 2006 Free Software Foundation, Inc.
+   2004, 2005, 2006, 2007 Free Software Foundation, Inc.
 
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
-   the Free Software Foundation; either version 2, or (at your option)
+   the Free Software Foundation; either version 3, or (at your option)
    any later version.
 
    This program is distributed in the hope that it will be useful,
@@ -386,9 +386,13 @@ extern int errno;
 #endif
 
 #ifndef ST_NBLOCKSIZE
-#define ST_NBLOCKSIZE 512
+# define ST_NBLOCKSIZE 512
 #endif
 
+#define ST_IS_SPARSE(st)                                  \
+  (ST_NBLOCKS (st)                                        \
+    < ((st).st_size / ST_NBLOCKSIZE + ((st).st_size % ST_NBLOCKSIZE != 0)))
+
 /* Declare standard functions.  */
 
 #if STDC_HEADERS
@@ -417,16 +421,7 @@ char *getenv ();
 # define MB_LEN_MAX 1
 #endif
 
-#if HAVE_INTTYPES_H
-# include <inttypes.h>
-#endif
-
-#if !HAVE_DECL_STRTOIMAX && !defined strtoimax
-intmax_t strtoimax ();
-#endif
-#if !HAVE_DECL_STRTOUMAX && !defined strtoumax
-uintmax_t strtoumax ();
-#endif
+#include <inttypes.h>
 
 #include <intprops.h>
 
@@ -458,6 +453,13 @@ uintmax_t strtoumax ();
 #define _(msgid) gettext (msgid)
 #define N_(msgid) msgid
 
+#ifdef HAVE_PWD_H
+# include <pwd.h>
+#endif
+#ifdef HAVE_GRP_H
+# include <grp.h>
+#endif
+
 #if MSDOS
 # include <process.h>
 # define SET_BINARY_MODE(arc) setmode(arc, O_BINARY)
@@ -466,13 +468,14 @@ uintmax_t strtoumax ();
 # define TTY_NAME "con"
 # define sys_reset_uid_gid()
 #else
-# include <pwd.h>
-# include <grp.h>
 # define SET_BINARY_MODE(arc)
 # define ERRNO_IS_EACCES 0
 # define TTY_NAME "/dev/tty"
-# define sys_reset_uid_gid() \
- do { setuid (getuid ()); setgid (getgid ()); } while (0)
+# define sys_reset_uid_gid()					\
+  do {								\
+    if (! (setuid (getuid ()) == 0 && setgid (getgid ()) == 0)) \
+      abort ();							\
+  } while (0)
 #endif
 
 #if XENIX
