@@ -1,7 +1,7 @@
 /* Various processing of names.
 
-   Copyright (C) 1988, 1992, 1994, 1996, 1997, 1998, 1999, 2000, 2001,
-   2003, 2004, 2005, 2006, 2007, 2009 Free Software Foundation, Inc.
+   Copyright 1988, 1992, 1994, 1996-2001, 2003-2007, 2009, 2013 Free
+   Software Foundation, Inc.
 
    This program is free software; you can redistribute it and/or modify it
    under the terms of the GNU General Public License as published by the
@@ -14,28 +14,20 @@
    Public License for more details.
 
    You should have received a copy of the GNU General Public License along
-   with this program; if not, write to the Free Software Foundation, Inc.,
-   51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.  */
+   with this program.  If not, see <http://www.gnu.org/licenses/>.  */
 
 #include <system.h>
 
 #include <fnmatch.h>
 #include <hash.h>
 #include <quotearg.h>
+#include <wordsplit.h>
+#include <argp.h>
 
 #include "common.h"
 
 /* User and group names.  */
 
-struct group *getgrnam ();
-struct passwd *getpwnam ();
-#if ! HAVE_DECL_GETPWUID
-struct passwd *getpwuid ();
-#endif
-#if ! HAVE_DECL_GETGRGID
-struct group *getgrgid ();
-#endif
-
 /* Make sure you link with the proper libraries if you are running the
    Yellow Peril (thanks for the good laugh, Ian J.!), or, euh... NIS.
    This code should also be modified for non-UNIX systems to do something
@@ -56,8 +48,6 @@ static char *cached_no_such_gname;
 static uid_t cached_no_such_uid;
 static gid_t cached_no_such_gid;
 
-static void register_individual_file (char const *name);
-
 /* Given UID, find the corresponding UNAME.  */
 void
 uid_to_uname (uid_t uid, char **uname)
@@ -179,7 +169,7 @@ gname_to_gid (char const *gname, gid_t *gidp)
 }
 
 
-struct name *
+static struct name *
 make_name (const char *file_name)
 {
   struct name *p = xzalloc (sizeof (*p));
@@ -190,7 +180,7 @@ make_name (const char *file_name)
   return p;
 }
 
-void
+static void
 free_name (struct name *p)
 {
   if (p)
@@ -205,51 +195,85 @@ free_name (struct name *p)
 /* Names from the command call.  */
 
 static struct name *namelist;	/* first name in list, if any */
-static struct name **nametail = &namelist;	/* end of name list */
+static struct name *nametail;	/* end of name list */
 
-/* File name arguments are processed in two stages: first a 
-   name_array (see below) is filled, then the names from it
+/* File name arguments are processed in two stages: first a
+   name element list (see below) is filled, then the names from it
    are moved into the namelist.
 
    This awkward process is needed only to implement --same-order option,
    which is meant to help process large archives on machines with
    limited memory.  With this option on, namelist contains at most one
    entry, which diminishes the memory consumption.
-   
+
    However, I very much doubt if we still need this -- Sergey */
 
-/* A name_array element contains entries of three types: */
+/* A name_list element contains entries of three types: */
 
 #define NELT_NAME  0   /* File name */
 #define NELT_CHDIR 1   /* Change directory request */
 #define NELT_FMASK 2   /* Change fnmatch options request */
+#define NELT_FILE  3   /* Read file names from that file */
+#define NELT_NOOP  4   /* No operation */
 
 struct name_elt        /* A name_array element. */
 {
+  struct name_elt *next, *prev;
   char type;           /* Element type, see NELT_* constants above */
   union
   {
     const char *name;  /* File or directory name */
-    int matching_flags;/* fnmatch options if type == NELT_FMASK */ 
+    int matching_flags;/* fnmatch options if type == NELT_FMASK */
+    struct             /* File, if type == NELT_FILE */
+    {
+      const char *name;/* File name */
+      int term;        /* File name terminator in the list */
+      FILE *fp;
+    } file;
   } v;
 };
 
-static struct name_elt *name_array;  /* store an array of names */
-static size_t allocated_names;	 /* how big is the array? */
-static size_t names;		 /* how many entries does it have? */
-static size_t name_index;	 /* how many of the entries have we scanned? */
+static struct name_elt *name_head;  /* store a list of names */
+size_t name_count;	 	    /* how many of the entries are names? */
 
-/* Check the size of name_array, reallocating it as necessary.  */
-static void
-check_name_alloc ()
+static struct name_elt *
+name_elt_alloc (void)
 {
-  if (names == allocated_names)
+  struct name_elt *elt;
+
+  elt = xmalloc (sizeof (*elt));
+  if (!name_head)
     {
-      if (allocated_names == 0)
-	allocated_names = 10; /* Set initial allocation */
-      name_array = x2nrealloc (name_array, &allocated_names,
-			       sizeof (name_array[0]));
+      name_head = elt;
+      name_head->prev = name_head->next = NULL;
+      name_head->type = NELT_NOOP;
+      elt = xmalloc (sizeof (*elt));
     }
+
+  elt->prev = name_head->prev;
+  if (name_head->prev)
+    name_head->prev->next = elt;
+  elt->next = name_head;
+  name_head->prev = elt;
+  return elt;
+}
+
+static void
+name_list_adjust (void)
+{
+  if (name_head)
+    while (name_head->prev)
+      name_head = name_head->prev;
+}
+
+static void
+name_list_advance (void)
+{
+  struct name_elt *elt = name_head;
+  name_head = elt->next;
+  if (name_head)
+    name_head->prev = NULL;
+  free (elt);
 }
 
 /* Add to name_array the file NAME with fnmatch options MATCHING_FLAGS */
@@ -257,33 +281,38 @@ void
 name_add_name (const char *name, int matching_flags)
 {
   static int prev_flags = 0; /* FIXME: Or EXCLUDE_ANCHORED? */
-  struct name_elt *ep;
+  struct name_elt *ep = name_elt_alloc ();
 
-  check_name_alloc ();
-  ep = &name_array[names++];
   if (prev_flags != matching_flags)
     {
       ep->type = NELT_FMASK;
       ep->v.matching_flags = matching_flags;
       prev_flags = matching_flags;
-      check_name_alloc ();
-      ep = &name_array[names++];
+      ep = name_elt_alloc ();
     }
   ep->type = NELT_NAME;
   ep->v.name = name;
+  name_count++;
 }
 
 /* Add to name_array a chdir request for the directory NAME */
 void
 name_add_dir (const char *name)
 {
-  struct name_elt *ep;
-  check_name_alloc ();
-  ep = &name_array[names++];
+  struct name_elt *ep = name_elt_alloc ();
   ep->type = NELT_CHDIR;
   ep->v.name = name;
-}  
+}
 
+void
+name_add_file (const char *name, int term)
+{
+  struct name_elt *ep = name_elt_alloc ();
+  ep->type = NELT_FILE;
+  ep->v.file.name = name;
+  ep->v.file.term = term;
+  ep->v.file.fp = NULL;
+}
 
 /* Names from external name file.  */
 
@@ -297,15 +326,222 @@ name_init (void)
 {
   name_buffer = xmalloc (NAME_FIELD_SIZE + 2);
   name_buffer_length = NAME_FIELD_SIZE;
+  name_list_adjust ();
 }
 
 void
 name_term (void)
 {
   free (name_buffer);
-  free (name_array);
+}
+
+/* Prevent recursive inclusion of the same file */
+struct file_id_list
+{
+  struct file_id_list *next;
+  ino_t ino;
+  dev_t dev;
+  const char *from_file;
+};
+
+static struct file_id_list *file_id_list;
+
+/* Return the name of the file from which the file names and options
+   are being read.
+*/
+static const char *
+file_list_name (void)
+{
+  struct name_elt *elt;
+
+  for (elt = name_head; elt; elt = elt->next)
+    if (elt->type == NELT_FILE && elt->v.file.fp)
+      return elt->v.file.name;
+  return _("command line");
 }
 
+static int
+add_file_id (const char *filename)
+{
+  struct file_id_list *p;
+  struct stat st;
+  const char *reading_from;
+
+  if (stat (filename, &st))
+    stat_fatal (filename);
+  reading_from = file_list_name ();
+  for (p = file_id_list; p; p = p->next)
+    if (p->ino == st.st_ino && p->dev == st.st_dev)
+      {
+	int oldc = set_char_quoting (NULL, ':', 1);
+	ERROR ((0, 0,
+		_("%s: file list requested from %s already read from %s"),
+		quotearg_n (0, filename),
+		reading_from, p->from_file));
+	set_char_quoting (NULL, ':', oldc);
+	return 1;
+      }
+  p = xmalloc (sizeof *p);
+  p->next = file_id_list;
+  p->ino = st.st_ino;
+  p->dev = st.st_dev;
+  p->from_file = reading_from;
+  file_id_list = p;
+  return 0;
+}
+
+enum read_file_list_state  /* Result of reading file name from the list file */
+  {
+    file_list_success,     /* OK, name read successfully */
+    file_list_end,         /* End of list file */
+    file_list_zero,        /* Zero separator encountered where it should not */
+    file_list_skip         /* Empty (zero-length) entry encountered, skip it */
+  };
+
+/* Read from FP a sequence of characters up to TERM and put them
+   into STK.
+ */
+static enum read_file_list_state
+read_name_from_file (struct name_elt *ent)
+{
+  int c;
+  size_t counter = 0;
+  FILE *fp = ent->v.file.fp;
+  int term = ent->v.file.term;
+
+  for (c = getc (fp); c != EOF && c != term; c = getc (fp))
+    {
+      if (counter == name_buffer_length)
+	name_buffer = x2realloc (name_buffer, &name_buffer_length);
+      name_buffer[counter++] = c;
+      if (c == 0)
+	{
+	  /* We have read a zero separator. The file possibly is
+	     zero-separated */
+	  return file_list_zero;
+	}
+    }
+
+  if (counter == 0 && c != EOF)
+    return file_list_skip;
+
+  if (counter == name_buffer_length)
+    name_buffer = x2realloc (name_buffer, &name_buffer_length);
+  name_buffer[counter] = 0;
+
+  return (counter == 0 && c == EOF) ? file_list_end : file_list_success;
+}
+
+static int
+handle_option (const char *str)
+{
+  struct wordsplit ws;
+  int i;
+
+  while (*str && isspace (*str))
+    ;
+  if (*str != '-')
+    return 1;
+
+  ws.ws_offs = 1;
+  if (wordsplit (str, &ws, WRDSF_DEFFLAGS|WRDSF_DOOFFS))
+    FATAL_ERROR ((0, 0, _("cannot split string '%s': %s"),
+		  str, wordsplit_strerror (&ws)));
+  ws.ws_wordv[0] = program_invocation_short_name;
+  more_options (ws.ws_wordc+ws.ws_offs, ws.ws_wordv);
+  for (i = 0; i < ws.ws_wordc+ws.ws_offs; i++)
+    ws.ws_wordv[i] = NULL;
+
+  wordsplit_free (&ws);
+  return 0;
+}
+
+static int
+read_next_name (struct name_elt *ent, struct name_elt *ret)
+{
+  if (!ent->v.file.fp)
+    {
+      if (!strcmp (ent->v.file.name, "-"))
+	{
+	  request_stdin ("-T");
+	  ent->v.file.fp = stdin;
+	}
+      else
+	{
+	  if (add_file_id (ent->v.file.name))
+	    {
+	      name_list_advance ();
+	      return 1;
+	    }
+	  if ((ent->v.file.fp = fopen (ent->v.file.name, "r")) == NULL)
+	    open_fatal (ent->v.file.name);
+	}
+    }
+
+  while (1)
+    {
+      switch (read_name_from_file (ent))
+	{
+	case file_list_skip:
+	  continue;
+
+	case file_list_zero:
+	  WARNOPT (WARN_FILENAME_WITH_NULS,
+		   (0, 0, N_("%s: file name read contains nul character"),
+		    quotearg_colon (ent->v.file.name)));
+	  ent->v.file.term = 0;
+	  /* fall through */
+	case file_list_success:
+	  if (handle_option (name_buffer) == 0)
+	    {
+	      name_list_adjust ();
+	      return 1;
+	    }
+	  ret->type = NELT_NAME;
+	  ret->v.name = name_buffer;
+	  return 0;
+
+	case file_list_end:
+	  if (strcmp (ent->v.file.name, "-"))
+	    fclose (ent->v.file.fp);
+	  ent->v.file.fp = NULL;
+	  name_list_advance ();
+	  return 1;
+	}
+    }
+}
+
+static void
+copy_name (struct name_elt *ep)
+{
+  const char *source;
+  size_t source_len;
+  char *cursor;
+
+  source = ep->v.name;
+  source_len = strlen (source);
+  if (name_buffer_length < source_len)
+    {
+      do
+	{
+	  name_buffer_length *= 2;
+	  if (! name_buffer_length)
+	    xalloc_die ();
+	}
+      while (name_buffer_length < source_len);
+
+      free (name_buffer);
+      name_buffer = xmalloc(name_buffer_length + 2);
+    }
+  strcpy (name_buffer, source);
+
+  /* Zap trailing slashes.  */
+  cursor = name_buffer + strlen (name_buffer) - 1;
+  while (cursor > name_buffer && ISSLASH (*cursor))
+    *cursor-- = '\0';
+}
+
+
 static int matching_flags; /* exclude_fnmatch options */
 
 /* Get the next NELT_NAME element from name_array.  Result is in
@@ -313,64 +549,48 @@ static int matching_flags; /* exclude_fnmatch options */
 
    If CHANGE_DIRS is true, treat any entries of type NELT_CHDIR as
    the request to change to the given directory.
-   
+
    Entries of type NELT_FMASK cause updates of the matching_flags
    value. */
-struct name_elt *
+static struct name_elt *
 name_next_elt (int change_dirs)
 {
   static struct name_elt entry;
-  const char *source;
-  char *cursor;
+  struct name_elt *ep;
 
-  while (name_index != names)
+  while ((ep = name_head) != NULL)
     {
-      struct name_elt *ep;
-      size_t source_len;
-      
-      ep = &name_array[name_index++];
-      if (ep->type == NELT_FMASK)
+      switch (ep->type)
 	{
+	case NELT_NOOP:
+	  name_list_advance ();
+	  break;
+
+	case NELT_FMASK:
 	  matching_flags = ep->v.matching_flags;
+	  name_list_advance ();
 	  continue;
-	}
-      
-      source = ep->v.name;
-      source_len = strlen (source);
-      if (name_buffer_length < source_len)
-	{
-	  do
+
+	case NELT_FILE:
+	  if (read_next_name (ep, &entry) == 0)
+	    return &entry;
+	  continue;
+
+	case NELT_CHDIR:
+	  if (change_dirs)
 	    {
-	      name_buffer_length *= 2;
-	      if (! name_buffer_length)
-		xalloc_die ();
+	      chdir_do (chdir_arg (xstrdup (ep->v.name)));
+	      name_list_advance ();
+	      break;
 	    }
-	  while (name_buffer_length < source_len);
-
-	  free (name_buffer);
-	  name_buffer = xmalloc (name_buffer_length + 2);
-	}
-      strcpy (name_buffer, source);
-
-      /* Zap trailing slashes.  */
-
-      cursor = name_buffer + strlen (name_buffer) - 1;
-      while (cursor > name_buffer && ISSLASH (*cursor))
-	*cursor-- = '\0';
-
-      if (change_dirs && ep->type == NELT_CHDIR)
-	{
-	  if (chdir (name_buffer) < 0)
-	    chdir_fatal (name_buffer);
-	}
-      else
-	{
+	  /* fall through */
+	case NELT_NAME:
+	  copy_name (ep);
 	  if (unquote_option)
 	    unquote_string (name_buffer);
-	  if (incremental_option)
-	    register_individual_file (name_buffer);
 	  entry.type = ep->type;
 	  entry.v.name = name_buffer;
+	  name_list_advance ();
 	  return &entry;
 	}
     }
@@ -421,9 +641,8 @@ name_gather (void)
 	  buffer->directory = NULL;
 	  buffer->parent = NULL;
 	  buffer->cmdline = true;
-	  
-	  namelist = buffer;
-	  nametail = &namelist->next;
+
+	  namelist = nametail = buffer;
 	}
       else if (change_dir)
 	addname (0, change_dir, false, NULL);
@@ -457,7 +676,7 @@ addname (char const *string, int change_dir, bool cmdline, struct name *parent)
 {
   struct name *name = make_name (string);
 
-  name->prev = *nametail;
+  name->prev = nametail;
   name->next = NULL;
   name->found_count = 0;
   name->matching_flags = matching_flags;
@@ -465,9 +684,12 @@ addname (char const *string, int change_dir, bool cmdline, struct name *parent)
   name->directory = NULL;
   name->parent = parent;
   name->cmdline = cmdline;
-  
-  *nametail = name;
-  nametail = &name->next;
+
+  if (nametail)
+    nametail->next = name;
+  else
+    namelist = name;
+  nametail = name;
   return name;
 }
 
@@ -501,7 +723,7 @@ remname (struct name *name)
   if ((p = name->next) != NULL)
     p->prev = name->prev;
   else
-    nametail = &name->prev;
+    nametail = name->prev;
 }
 
 /* Return true if and only if name FILE_NAME (from an archive) matches any
@@ -517,12 +739,12 @@ name_match (const char *file_name)
 
       if (!cursor)
 	return true;
-      
+
       if (cursor->name[0] == 0)
 	{
 	  chdir_do (cursor->change_dir);
-	  namelist = 0;
-	  nametail = &namelist;
+	  namelist = NULL;
+	  nametail = NULL;
 	  return true;
 	}
 
@@ -535,8 +757,8 @@ name_match (const char *file_name)
 	  if (starting_file_option)
 	    {
 	      free (namelist);
-	      namelist = 0;
-	      nametail = &namelist;
+	      namelist = NULL;
+	      nametail = NULL;
 	    }
 	  chdir_do (cursor->change_dir);
 
@@ -575,8 +797,6 @@ all_names_found (struct tar_stat_info *p)
   struct name const *cursor;
   size_t len;
 
-  if (test_label_option)
-    return true;
   if (!p->file_name || occurrence_option == 0 || p->had_trailing_slash)
     return false;
   len = strlen (p->file_name);
@@ -589,29 +809,21 @@ all_names_found (struct tar_stat_info *p)
   return true;
 }
 
-static inline int
-is_pattern (const char *string)
-{
-  return strchr (string, '*') || strchr (string, '[') || strchr (string, '?');
-}
-
-static void
+static int
 regex_usage_warning (const char *name)
 {
   static int warned_once = 0;
 
-  if (warn_regex_usage && is_pattern (name))
+  if (warn_regex_usage && fnmatch_pattern_has_wildcards (name, 0))
     {
       warned_once = 1;
       WARN ((0, 0,
-	     /* TRANSLATORS: The following three msgids form a single sentence.
-	      */
-	     _("Pattern matching characters used in file names. Please,")));
+	     _("Pattern matching characters used in file names")));
       WARN ((0, 0,
-	     _("use --wildcards to enable pattern matching, or --no-wildcards to")));
-      WARN ((0, 0,
-	     _("suppress this warning.")));
+	     _("Use --wildcards to enable pattern matching,"
+	       " or --no-wildcards to suppress this warning")));
     }
+  return warned_once;
 }
 
 /* Print the names of things in the namelist that were not matched.  */
@@ -624,17 +836,16 @@ names_notfound (void)
     if (!WASFOUND (cursor) && cursor->name[0])
       {
 	regex_usage_warning (cursor->name);
-	if (cursor->found_count == 0)
-	  ERROR ((0, 0, _("%s: Not found in archive"),
-		  quotearg_colon (cursor->name)));
-	else
-	  ERROR ((0, 0, _("%s: Required occurrence not found in archive"),
-		  quotearg_colon (cursor->name)));  
+	ERROR ((0, 0,
+		(cursor->found_count == 0) ?
+		     _("%s: Not found in archive") :
+		     _("%s: Required occurrence not found in archive"),
+		quotearg_colon (cursor->name)));
       }
 
   /* Don't bother freeing the name list; we're about to exit.  */
-  namelist = 0;
-  nametail = &namelist;
+  namelist = NULL;
+  nametail = NULL;
 
   if (same_order_option)
     {
@@ -648,14 +859,50 @@ names_notfound (void)
 	}
     }
 }
+
+void
+label_notfound (void)
+{
+  struct name const *cursor;
+
+  if (!namelist)
+    return;
+
+  for (cursor = namelist; cursor; cursor = cursor->next)
+    if (WASFOUND (cursor))
+      return;
+
+  if (verbose_option)
+    error (0, 0, _("Archive label mismatch"));
+  set_exit_status (TAREXIT_DIFFERS);
+
+  for (cursor = namelist; cursor; cursor = cursor->next)
+    {
+      if (regex_usage_warning (cursor->name))
+	break;
+    }
+
+  /* Don't bother freeing the name list; we're about to exit.  */
+  namelist = NULL;
+  nametail = NULL;
+
+  if (same_order_option)
+    {
+      const char *name;
+
+      while ((name = name_next (1)) != NULL
+	     && regex_usage_warning (name) == 0)
+	;
+    }
+}
 
 /* Sorting name lists.  */
 
 /* Sort *singly* linked LIST of names, of given LENGTH, using COMPARE
    to order names.  Return the sorted list.  Note that after calling
-   this function, the `prev' links in list elements are messed up.
-   
-   Apart from the type `struct name' and the definition of SUCCESSOR,
+   this function, the 'prev' links in list elements are messed up.
+
+   Apart from the type 'struct name' and the definition of SUCCESSOR,
    this is a generic list-sorting function, but it's too painful to
    make it both generic and portable
    in C.  */
@@ -761,17 +1008,15 @@ compare_names (struct name const *n1, struct name const *n2)
 }
 
 
-/* Add all the dirs under NAME, which names a directory, to the namelist.
-   If any of the files is a directory, recurse on the subdirectory.
-   DEVICE is the device not to leave, if the -l option is specified.
-   CMDLINE is true, if the NAME appeared on the command line. */
+/* Add all the dirs under ST to the namelist NAME, descending the
+   directory hierarchy recursively.  */
 
 static void
-add_hierarchy_to_namelist (struct name *name, dev_t device, bool cmdline)
+add_hierarchy_to_namelist (struct tar_stat_info *st, struct name *name)
 {
   const char *buffer;
-  
-  name_fill_directory (name, device, cmdline);
+
+  name->directory = scan_directory (st);
   buffer = directory_contents (name->directory);
   if (buffer)
     {
@@ -799,6 +1044,8 @@ add_hierarchy_to_namelist (struct name *name, dev_t device, bool cmdline)
 	  if (*string == 'D')
 	    {
 	      struct name *np;
+	      struct tar_stat_info subdir;
+	      int subfd;
 
 	      if (allocated_length <= name_length + string_length)
 		{
@@ -819,7 +1066,38 @@ add_hierarchy_to_namelist (struct name *name, dev_t device, bool cmdline)
 	      else
 		child_tail->sibling = np;
 	      child_tail = np;
-	      add_hierarchy_to_namelist (np, device, false);
+
+	      tar_stat_init (&subdir);
+	      subdir.parent = st;
+	      if (st->fd < 0)
+		{
+		  subfd = -1;
+		  errno = - st->fd;
+		}
+	      else
+		subfd = subfile_open (st, string + 1,
+				      open_read_flags | O_DIRECTORY);
+	      if (subfd < 0)
+		open_diag (namebuf);
+	      else
+		{
+		  subdir.fd = subfd;
+		  if (fstat (subfd, &subdir.stat) != 0)
+		    stat_diag (namebuf);
+		  else if (! (O_DIRECTORY || S_ISDIR (subdir.stat.st_mode)))
+		    {
+		      errno = ENOTDIR;
+		      open_diag (namebuf);
+		    }
+		  else
+		    {
+		      subdir.orig_file_name = xstrdup (namebuf);
+		      add_hierarchy_to_namelist (&subdir, np);
+		      restore_parent_fd (&subdir);
+		    }
+		}
+
+	      tar_stat_destroy (&subdir);
 	    }
 	}
 
@@ -847,7 +1125,7 @@ name_compare (void const *entry1, void const *entry2)
 }
 
 
-/* Rebase `name' member of CHILD and all its siblings to
+/* Rebase 'name' member of CHILD and all its siblings to
    the new PARENT. */
 static void
 rebase_child_list (struct name *child, struct name *parent)
@@ -855,7 +1133,7 @@ rebase_child_list (struct name *child, struct name *parent)
   size_t old_prefix_len = child->parent->length;
   size_t new_prefix_len = parent->length;
   char *new_prefix = parent->name;
-  
+
   for (; child; child = child->sibling)
     {
       size_t size = child->length - old_prefix_len + new_prefix_len;
@@ -867,7 +1145,7 @@ rebase_child_list (struct name *child, struct name *parent)
       child->length = size;
 
       rebase_directory (child->directory,
-			child->parent->name, old_prefix_len, 
+			child->parent->name, old_prefix_len,
 			new_prefix, new_prefix_len);
     }
 }
@@ -880,11 +1158,10 @@ void
 collect_and_sort_names (void)
 {
   struct name *name;
-  struct name *next_name, *prev_name;
+  struct name *next_name, *prev_name = NULL;
   int num_names;
-  struct stat statbuf;
   Hash_table *nametab;
-  
+
   name_gather ();
 
   if (!namelist)
@@ -909,13 +1186,15 @@ collect_and_sort_names (void)
 			_("Only one -C option is allowed with "
 			  "--listed-incremental")));
 	}
-      chdir_do (namelist->change_dir);
+
       read_directory_file ();
     }
-  
+
   num_names = 0;
   for (name = namelist; name; name = name->next, num_names++)
     {
+      struct tar_stat_info st;
+
       if (name->found_count || name->directory)
 	continue;
       if (name->matching_flags & EXCLUDE_WILDCARDS)
@@ -927,28 +1206,44 @@ collect_and_sort_names (void)
       if (name->name[0] == 0)
 	continue;
 
-      if (deref_stat (dereference_option, name->name, &statbuf) != 0)
+      tar_stat_init (&st);
+
+      if (deref_stat (name->name, &st.stat) != 0)
 	{
 	  stat_diag (name->name);
 	  continue;
 	}
-      if (S_ISDIR (statbuf.st_mode))
+      if (S_ISDIR (st.stat.st_mode))
 	{
-	  name->found_count++;
-	  add_hierarchy_to_namelist (name, statbuf.st_dev, true);
+	  int dir_fd = openat (chdir_fd, name->name,
+			       open_read_flags | O_DIRECTORY);
+	  if (dir_fd < 0)
+	    open_diag (name->name);
+	  else
+	    {
+	      st.fd = dir_fd;
+	      if (fstat (dir_fd, &st.stat) != 0)
+		stat_diag (name->name);
+	      else if (O_DIRECTORY || S_ISDIR (st.stat.st_mode))
+		{
+		  st.orig_file_name = xstrdup (name->name);
+		  name->found_count++;
+		  add_hierarchy_to_namelist (&st, name);
+		}
+	    }
 	}
+
+      tar_stat_destroy (&st);
     }
 
   namelist = merge_sort (namelist, num_names, compare_names);
 
   num_names = 0;
-  nametab = hash_initialize (0, 0,
-			     name_hash,
-			     name_compare, NULL);
+  nametab = hash_initialize (0, 0, name_hash, name_compare, NULL);
   for (name = namelist; name; name = next_name)
     {
       next_name = name->next;
-      name->caname = normalize_filename (name->name);
+      name->caname = normalize_filename (name->change_dir, name->name);
       if (prev_name)
 	{
 	  struct name *p = hash_lookup (nametab, name);
@@ -959,6 +1254,7 @@ collect_and_sort_names (void)
 		{
 		  if (p->child)
 		    rebase_child_list (p->child, name);
+		  hash_delete (nametab, name);
 		  /* FIXME: remove_directory (p->caname); ? */
 		  remname (p);
 		  free_name (p);
@@ -981,7 +1277,7 @@ collect_and_sort_names (void)
       prev_name = name;
       num_names++;
     }
-  nametail = &prev_name;
+  nametail = prev_name;
   hash_free (nametab);
 
   namelist = merge_sort (namelist, num_names, compare_names_found);
@@ -1033,7 +1329,7 @@ name_scan (const char *file_name)
 struct name *gnu_list_name;
 
 struct name const *
-name_from_list ()
+name_from_list (void)
 {
   if (!gnu_list_name)
     gnu_list_name = namelist;
@@ -1080,46 +1376,6 @@ excluded_name (char const *name)
 {
   return excluded_file_name (excluded, name + FILE_SYSTEM_PREFIX_LEN (name));
 }
-
-/* Names to avoid dumping.  */
-static Hash_table *avoided_name_table;
-
-/* Remember to not archive NAME.  */
-void
-add_avoided_name (char const *name)
-{
-  hash_string_insert (&avoided_name_table, name);
-}
-
-/* Should NAME be avoided when archiving?  */
-bool
-is_avoided_name (char const *name)
-{
-  return hash_string_lookup (avoided_name_table, name);
-}
-
-
-static Hash_table *individual_file_table;
-
-static void
-register_individual_file (char const *name)
-{
-  struct stat st;
-  
-  if (deref_stat (dereference_option, name, &st) != 0)
-    return; /* Will be complained about later */
-  if (S_ISDIR (st.st_mode))
-    return;
-  
-  hash_string_insert (&individual_file_table, name);
-}
-
-bool
-is_individual_file (char const *name)
-{
-  return hash_string_lookup (individual_file_table, name);
-}
-
 
 
 /* Return the size of the prefix of FILE_NAME that is removed after
